local HeadCount = LibStub("AceAddon-3.0"):GetAddon("HeadCount2")
local Tracker = HeadCount:NewModule("RaidTracker","AceEvent-3.0","AceTimer-3.0")
if not HeadCount then return end

local L = LibStub("AceLocale-3.0"):GetLocale("HeadCount2")

local InstanceBossList = {}

function Tracker:OnInitialize()
	-- Setup Event Handlers
	self.isEnabled = true
	self.isWaitlistAcceptanceEnabled = false
	self.isCurrentRaidActive = false
	self.currentBossEvent = HeadCount:NewInstance("BossEvent")
	self.raidListWrapper = HeadCount:NewInstance("RaidListWrapper")
	self.messageHandler = HeadCount:GetModule("MessageHandler")
end

function Tracker:GetInstaceBossList()
	local lockTimeleft, isPreviousInstance, encountersTotal, encountersComplete = GetInstanceLockTimeRemaining();
	for i=1,encountersTotal do
		bossName, texture, isKilled = GetInstanceLockTimeRemainingEncounter(i);
		InstanceBossList[bossName] = i
	end
end

function Tracker:SetRaidListWrapper(wrapper)
	if wrapper then
		self.raidListWrapper = wrapper
	end
end

function Tracker:saveRaidListWrapper()
	HeadCount.db.profile.raidListWrapper = self.raidListWrapper
end

function Tracker:recover()
	local raid = self:retrieveMostRecentRaid()
	if raid then
		local ts = HeadCount:getUTCDateTimeInSeconds()
		local activityTime = HeadCount:NewInstance("Time",ts)
		if UnitInRaid("player")	 then
			-- owner is present in raid
			if raid.isFinalized  then
				-- most recent raid is finalized
				HeadCount:LogDebug(string.format(L["debug.raidtracker.setup.raid.final"], HeadCount.TITLE, HeadCount.VERSION))
				self.isCurrentRaidActive = false
			else
				-- most recent raid is NOT finalized
				HeadCount:LogDebug(string.format(L["debug.raidtracker.setup.raid.nofinal"], HeadCount.TITLE, HeadCount.VERSION))
				raid:recover(activityTime)
				-- Set proper events and flags for recovered raid
				self:beginRaid()
			end
		else
			-- owner is NOT present in raid
			if raid.isFinalized  then
				-- most recent raid is finalized
				HeadCount:LogDebug(string.format(L["debug.raidtracker.setup.noraid.final"], HeadCount.TITLE, HeadCount.VERSION))
				self.isCurrentRaidActive = false
			else
				-- most recent raid is NOT finalized
				HeadCount:LogDebug(string.format(L["debug.raidtracker.setup.noraid.nofinal"], HeadCount.TITLE, HeadCount.VERSION))
				self:endRaid()
				raid:finalize(activityTime)
				self.isCurrentRaidActive = false
			end
		end
	else
		--HeadCount:LogInformation("The most recent raid does not exist.")
	end
end

function Tracker:CHAT_MSG_WHISPER_FILTER(chatFrame,event,...)
	local message = ...
	if messageHandler:IsSupportedIncomingMessage(message) then
		-- waitlist message and author was not self
		if not HeadCount.db.profile.raid["waitlist"]["incoming"] then
			-- waitlist incoming message display is disabled
			return true, ...
		end
	end
end
function Tracker:CHAT_MSG_WHISPER_INFORM_FILTER(chatFrame,event,...)
	local message = ...
	if messageHandler:IsSupportedOutgoingMessage(message) and not HeadCount.db.profile.raid["waitlist"]["incoming"] then
		-- outgoing message is a waitlist message and outgoing waitlist message display is disabled
		return true, ...
	end
end


function Tracker:beginRaid()
	self.isCurrentRaidActive = true
	local waitlistDuration = HeadCount.db.profile.raid.waitlist.duration
	if waitlistDuration > 0 then
		self.isWaitlistAcceptanceEnabled = false
	else
		self.isWaitlistAcceptanceEnabled = true
	end
	ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", Tracker.CHAT_MSG_WHISPER_FILTER)			-- enable received whisper event tracking
	ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", Tracker.CHAT_MSG_WHISPER_INFORM_FILTER)	-- enable sent whisper event tracking
	self:RegisterEvent("CHAT_MSG_WHISPER")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")				-- enable being added to combat event tracking for boss kill processing
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")			-- disable being removed from combat event tracking
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")	-- disable combat log event tracking
	self:UnregisterEvent("CHAT_MSG_MONSTER_YELL")			-- disable monster yell events
end

-- Processes ending of a raid.
function Tracker:endRaid()
	self.isCurrentRaidActive = false						-- current raid is inactive
	self.isWaitlistAcceptanceEnabled = false				-- waitlist acceptance is disabled
	ChatFrame_RemoveMessageEventFilter("CHAT_MSG_WHISPER", Tracker.CHAT_MSG_WHISPER_FILTER)			-- enable received whisper event tracking
	ChatFrame_RemoveMessageEventFilter("CHAT_MSG_WHISPER_INFORM", Tracker.CHAT_MSG_WHISPER_INFORM_FILTER)	-- enable sent whisper event tracking
	self:UnregisterEvent("CHAT_MSG_WHISPER")
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:UnregisterEvent("CHAT_MSG_MONSTER_YELL")
	self.currentBossEvent:Stop()
end

-- Determines if there is an active raid.
-- @return boolean Returns true if there is an active raid, and false otherwise.
function Tracker:isRaidActive()
	return self.isCurrentRaidActive and self.raidListWrapper.numberOfRaids > 0
end
function Tracker:endCurrentRaid(activityTime)
	if activityTime and self.isCurrentRaidActive then
		-- end the currently active raid
		if self.raidListWrapper:retrieveMostRecentRaid() then
			self:endRaid()
			-- finalize the current raid
			currentRaid:finalize(activityTime)
			return true
		end
	end
	return false
end

function Tracker:removeRaid(id)
	local isRaidRemoved = false
	local isCurrentRaidEnded = false

	local raid = self.raidListWrapper:getRaidById(id)
	if raid then
		-- raid is present and raid
		if not raid.isFinalized and self.isCurrentRaidActive then
			-- raid is not finalized, raid is active
			self:endRaid()
			isCurrentRaidEnded = true
		end
		self.raidListWrapper:removeRaid(id)
		isRaidRemoved = true
	end
	return isRaidRemoved, isCurrentRaidEnded
end
function Tracker:removeAllRaids()
	self:endRaid()
	self.raidListWrapper:removeAll()
end
function Tracker:removePlayer(id, playerName)
	local raid = self.raidListWrapper:getRaidById(id)
	if raid then raid:removePlayer(playerName) end
end
function Tracker:removeWaitlistPlayer(id, playerName)
	local raid = self.raidListWrapper:getRaidById(id)
	if raid then raid:removeWaitlistPlayer(playerName) end
end
function Tracker:removeBoss(id, bossName)
	local raid = self.raidListWrapper:getRaidById(id)
	if raid then raid:removeBoss(bossName) end
end
function Tracker:removeEventAttendee(raidId, bossName, attendeeId)
	local raid = self.raidListWrapper:getRaidById(raidId)
	if raid then raid:removeEventAttendee(bossName, attendeeId)	end
end

function Tracker:removeLoot(raidId, lootId)
	local raid = self.raidListWrapper:getRaidById(raidId)
	if raid then raid:removeLoot(lootId) end
end
function Tracker:addRaid(raid)
	local raidStartTime = raid:retrieveStartingTime().epoc
	self.raidListWrapper:addRaid(raidStartTime, raid)
end
function Tracker:retrieveCurrentRaid(activityTime)
	local currentRaid = nil
	local isNewRaid = false
	if self.isCurrentRaidActive then
		-- a raid currently exists, process existing raid
		currentRaid = self.raidListWrapper:retrieveMostRecentRaid()
		isNewRaid = false
	else
		-- a raid does not already exist, create a new raid, then process it
		local args = {
			["playerList"] = {
				["raidgroup"] = { },
				["nogroup"] = { },
			},
			["bossList"] = { },
			["lootList"] = { },
			["timeList"] = { },
			["difficulty"] = nil,
			["raidTime"] = 0,
			["lastActivityTime"] = activityTime,
			["zone"] = nil,
			["numberOfPlayers"] = 0,
			["numberOfBosses"] = 0,
			["isFinalized"] = false
		}

		local timePair = HeadCount:NewInstance("TimePair",{ ["start"] = activityTime})
		table.insert(args["timeList"], timePair)	-- insert the first time pair (includes begin time with no end time)
		currentRaid = HeadCount:NewInstance("Raid",args)
		self:addRaid(currentRaid)
		isNewRaid = true
		HeadCount:LogInformation(string.format(L["info.newraid"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(currentRaid:getLastActivityTime())))
		-- Set proper events and flags for new raid
		self:beginRaid()
	end
	return currentRaid, isNewRaid
end
function Tracker:processStatus(zone)
	assert(zone, "Unable to determine battleground status because the zone is nil.")
	local inInstance, instanceType = IsInInstance()
	if not HeadCount.db.profile.isBattlegroundTrackingEnabled and HeadCount:isBattlegroundZone(zone) then
		-- battleground tracking is disabled and current zone is a battleground
		if self.isEnabled then
			HeadCount:LogDebug(string.format(L["debug.status.battleground.enter"], HeadCount.TITLE, HeadCount.VERSION, zone))
			self.isEnabled = false
		end
	elseif not self.isEnabled then
		if instanceType == "raid" then
			self.isEnabled = true
		end
	end
end


function Tracker:processAutomaticGroupSelection(zone)
	local isAutomaticGroupSelectionEnabled = HeadCount.db.profile.groupsetup.isAutomaticGroupSelectionEnabled
	local isPresentInInstance, instanceType = IsInInstance()
	if isPresentInInstance and instanceType == "raid" and self.isCurrentRaidActive and isAutomaticGroupSelectionEnabled then
		local name, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic = GetInstanceInfo()
		if type == "raid" then
			local numberOfPlayers = maxPlayers
			local numberOfRaidGroups = numberOfPlayers / HeadCount.NUMBER_OF_PARTY_PLAYERS
			for i = 1, HeadCount.NUMBER_OF_PARTY_GROUPS do
				-- 10 / 5 = 2,  20 / 5 = 4,  25 / 5 = 5, 40 / 5 = 8
				local partyGroupString = string.format("%d", i)
				if (i <= numberOfRaidGroups) then
					self.db.profile.raidListGroups[partyGroupString] = true
					self.db.profile.waitListGroups[partyGroupString] = false
				else
					self.db.profile.raidListGroups[partyGroupString] = false
					self.db.profile.waitListGroups[partyGroupString] = true
				end
			end
		end
	end
end

-- Processes the current raid zone.
-- If the current raid has no raid zone assigned to it, set the raid zone name.
-- @param zone The zone name.
function Tracker:processRaidZone(zone)
	local isPresentInInstance, instanceType = IsInInstance()
	if isPresentInInstance and instanceType == "raid" and self.isCurrentRaidActive then
		-- player is in a raid instance and a tracked raid is active
		local currentRaid = self:retrieveMostRecentRaid()
		local name, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic = GetInstanceInfo()
		if not currentRaid.zone or currentRaid.zone == UNKNOWN  then
			-- current raid zone is not set
			local localizedZoneName = name
			currentRaid.zone = localizedZoneName
			self:saveRaidListWrapper()	-- save the raid list
		end
	end
end

-- Processes the difficulty
function Tracker:processDifficulty()
	local isPresentInInstance, instanceType = IsInInstance()
	if isPresentInInstance and instanceType == "raid" and self.isCurrentRaidActive then
		-- player is in a raid instance and a tracked raid is active
		local currentRaid = self.raidListWrapper:retrieveMostRecentRaid()
		if not currentRaid.difficulty then
			local difficulty = GetDifficulty()
			local name, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic = GetInstanceInfo()
			if difficulty then
				HeadCount:LogDebug(string.format(L["debug.raid.update.difficulty"], HeadCount.TITLE, HeadCount.VERSION, difficultyName))
				currentRaid.difficulty = difficulty
				self:saveRaidListWrapper()	-- save the raid list
			end
		end
	end
end


function Tracker:processLootUpdate(message)
	local isProcessed = false

	if self.isCurrentRaidActive then
		-- current raid is active
		local playerName, item, quantity = self:processLootPattern(message)
		if playerName then
			-- message matched proper loot pattern
			local utcDateTimeInSeconds = HeadCount:getUTCDateTimeInSeconds()
			local activityTime = HeadCount:NewInstance("Time",utcDateTimeInSeconds)

			-- Process the loot source if it is a new boss encounter
			local isBoss, lootSource = self:processLootSource()
			local encounterName = HeadCount:retrieveBossEncounterName(lootSource)
			local isBossKillProcessed = false
			if isBoss then
				-- loot source is valid and is a raid boss, process it for possible boss kill
				local zone = GetRealZoneText()
				local difficulty = HeadCount:determineDifficulty()
				isBossKillProcessed = self:addBossKill(encounterName, zone, difficulty, activityTime)
			end
			-- Process the loot
			local isLootProcessed = self:processLoot(playerName, item, quantity, activityTime, encounterName)
			isProcessed = (isBossKillProcessed or isLootProcessed)
		end
	end

	return isProcessed
end

-- Process the loot source
-- @return isBoss Returns if the loot source is a boss
-- @return lootSource Returns the loot source (mob name)
function Tracker:processLootSource()
	local isBoss = false
	local lootSource = HeadCount.DEFAULT_LOOT_SOURCE

	local numberOfRaidMembers = GetNumRaidMembers()
	for i = 1, numberOfRaidMembers do
		-- Raid target
		local unit = "raid" .. i .. "target"
		if (self:isBossUnitDead(unit)) then
			isBoss = true
			lootSource = UnitName(unit)
			break
		end
	end

	return isBoss, lootSource
end

-- Processing a loot
-- @param playerName The player name.
-- @param item The item string
-- @param quantity The loot quantity
-- @param activityTime The activity time.
-- @param source The loot source
-- @return boolean Returns true if loot was processed and false otherwise
function Tracker:processLoot(playerName, item, quantity, activityTime, source)
	local isProcessed = false
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(item)

	local minimumLootQuality = HeadCount.db.profile.minimumLootQuality
	local id = string.match(itemLink, "item:(%d+):")
	local numberId = tonumber(id)
	local exclusionList = HeadCount:GetExclusionList()
	if itemRarity >= minimumLootQuality and not exclusionList[numberId] then
		-- item dropped is greater than or equal to minimum loot quality and the item is not excluded, track it!
		local realZoneText = GetRealZoneText()
		local args = {
			["player"] = playerName,
			["cost"] = 0,
			["activityTime"] = activityTime,
			["quantity"] = quantity,
			["zone"] = realZoneText,
			["source"] = source,
			["link"] = itemLink,
			["note"] = nil
		}
		--local itemId = AceLibrary("HeadCountLoot-1.0"):retrieveItemId(itemLink)	-- get the item id
		local currentRaid = self:retrieveMostRecentRaid()
		local loot = HeadCount:NewInstance("Loot", args)

		-- Add this loot to the raid
		currentRaid:addLoot(loot)

		-- Should we broadcast this loot to guild chat?
		if HeadCount.db.profile.isLootBroadcastEnabled then
			-- player is in a guild and loot broadcasting is enabled, broadcast to the guild!
			local channel = HeadCount.db.profile.broadcastChannel
			HeadCount:SendMessage(string.format(L["guild.broadcast.loot"], playerName, itemLink, source), channel)
		end

		-- Should the loot management popup window show?
		if HeadCount.db.profile.isLootPopupEnabled then
			-- loot popup window should show
			local raidId = currentRaid:retrieveStartingTime().epoc
			local lootId = #currentRaid.lootList
			HeadCount:GetModule("GUI"):ShowLootPopup(raidId, lootId)
		end
		isProcessed = true
	end
	return isProcessed
end

-- Processes a loot message pattern
-- @param message The loot message.
-- @return string, string, number  Returns the player name, item name, and quantity or nil if the message does not match any pattern
function Tracker:processLootPattern(message)
	local playerName = nil
	local item = nil
	local quantity = nil

	local multipleLootSelfRegex = HeadCount:getMULTIPLE_LOOT_SELF_REGEX()
	local lootSelfRegex = HeadCount:getLOOT_SELF_REGEX()
	local multipleLootRegex = HeadCount:getMULTIPLE_LOOT_REGEX()
	local lootRegex = HeadCount:getLOOT_REGEX()

	if (message:match(multipleLootSelfRegex)) then
		item, quantity = message:match(multipleLootSelfRegex)
		playerName = UnitName("player")
	elseif (message:match(lootSelfRegex)) then
		item = message:match(lootSelfRegex)
		playerName = UnitName("player")
		quantity = 1
	elseif (message:match(multipleLootRegex)) then
		playerName, item, quantity = message:match(multipleLootRegex)
	elseif (message:match(lootRegex)) then
		playerName, item = message:match(lootRegex)
		quantity = 1
	end
	return playerName, item, quantity
end

-- Add a boss kill.
-- @param encounterName The boss encounter name.
-- @param zone The zone name.
-- @param activityTime The boss kill time.
-- @return boolean Returns true if a boss kill was processed and false otherwise
function Tracker:addBossKill(encounterName, zone, difficulty, activityTime)
	local isProcessed = false
	local currentRaid = self:retrieveMostRecentRaid()
	if not currentRaid:isBossPresent(encounterName) then
		HeadCount:LogDebug(string.format(L["debug.boss.kill.complete"], HeadCount.TITLE, HeadCount.VERSION, encounterName))

		self:processAttendance(activityTime)	-- update attendance

		local raidMemberSort = HeadCount:getRAID_MEMBER_SORT()
		local playerList = currentRaid:retrieveOrderedPlayerList("Name",
			HeadCount.db.profile.boss["groupsetup"]["raidlistgroup"], HeadCount.db.profile.boss["groupsetup"]["waitlistgroup"],	HeadCount.db.profile.boss["groupsetup"]["nolistgroup"],
			HeadCount.db.profile.boss["groupsetup"]["waitlist"], true, true)

		local args = {
			["name"] = encounterName,
			["zone"] = zone,
			["difficulty"] = difficulty,
			["runtime"] = activityTime,
			["playerList"] = playerList
		}
		local boss = HeadCount:NewInstance("Boss",args)
		currentRaid:addBoss(boss)
		-- Should we broadcast this to guild chat?
		if HeadCount.db.profile.isLootBroadcastEnabled then
			-- boss kill broadcasting is enabled
			local channel =  HeadCount.db.profile.broadcastChannel
			HeadCount:SendMessage(string.format(L["guild.broadcast.bosskill"], encounterName), channel)
		end
		isProcessed = true
	end
	return isProcessed
end

-- Determines if the given unit is dead and a boss.
-- @param unit The unit.
-- @return boolean Returns true if the unit is a dead boss and false otherwise.
function Tracker.fixme:isBossUnitDead(unit)
	local isValid = false

	if unit then
		if UnitExists(unit) and UnitIsDead(unit) then
			-- unit exists and unit is dead
			local unitName = UnitName(unit)
			-- check SavedInstance Info HERE

			-- @TODO FIX ME

			if UnitClassification(unit) == "worldboss" and not HeadCount.BOSS_BLACKLIST[unitName] then
				isValid = true
			elseif HeadCount.BOSS_WHITELIST[unitName] then
				isValid = true
			end
		end
	end

	return isValid
end

-------------------------------
-- ATTENDANCE TRACKING
-------------------------------
-- Process an attendance update
-- @param activityTime The current activity time.
function Tracker:processAttendance(activityTime)
	assert(activityTime, "Unable to update attendance because the activity time is nil.")

	-- Get the current zone
	local zone = GetRealZoneText()
	-- update the status
	self:processStatus(zone)

	if self.isEnabled then
		if UnitInRaid("player") then
			-- player is in a raid
			local currentRaid, isNewRaid = self:retrieveCurrentRaid(activityTime)	-- retrieve the current raid or create a new raid
			if isNewRaid then
				-- new raid
				--HeadCount:DisableModalFrame() 	-- disable all modal frames

				if UnitAffectingCombat("player") then
					self:PLAYER_REGEN_DISABLED()	-- call the combat event immediately
				end
			end

			if not currentRaid.zone or currentRaid.zone == UNKNONW then
				-- process automatic group selection if enabled and only if no zone is already defined for this raid
				self:processAutomaticGroupSelection(zone)
			end
			self:processRaidZone(zone)			-- Assign zone if current raid does not have a zone assigned
			self:processDifficulty()			-- Assign difficulty is current difficulty is not yet assigned

			-- Update each member in the raid group
			local numberOfRaidMembers = GetNumRaidMembers()
			if (numberOfRaidMembers > 0) then
				-- active in a raid

				local trackedPlayerList = { }
				for i = 1,numberOfRaidMembers do
					local name, rank, subgroup, level, className, fileName, zone, isOnline, isDead, role, isML = GetRaidRosterInfo(i)

					-- self:debugRosterMember(name, rank, subgroup, level, className, fileName, zone, isOnline, isDead, role, isML)
					if currentRaid:isPlayerPresent(name) then
						-- player is already present in the player list (existing player)
						local player = currentRaid:retrievePlayer(name)
						local timeDifference = HeadCount:computeTimeDifference(activityTime, player.lastActivityTime)
						local moveResult

						if (isOnline) then
							-- CURRENT: Player is online
							local isPresentInRaidList = HeadCount.db.profile.raidListGroups[subgroup]
							local isPresentInWaitList = HeadCount.db.profile.waitListGroups[subgroup]

							if isPresentInRaidList then
								-- CURRENT: Player is in the raid list.
								moveResult = player:moveToRaidList(activityTime)
								if moveResult then
									currentRaid:moveToRaidGroup(player)
									HeadCount:LogDebug(string.format(L["debug.raid.update.modifymember.raidlist"], HeadCount.TITLE, HeadCount.VERSION, name))
								end
							elseif isPresentInWaitList then
								-- CURRENT: Player is in the wait list.
								moveResult = player:moveToWaitList(activityTime)
								if moveResult then
									currentRaid:moveToRaidGroup(player)
									HeadCount:LogDebug(string.format(L["debug.raid.update.modifymember.waitlist"], HeadCount.TITLE, HeadCount.VERSION, name))
								end
							else
								moveResult = player:moveToNoList(activityTime, true)
								if moveResult then
									currentRaid:moveToNoGroup(player)
									HeadCount:LogDebug(string.format(L["debug.raid.update.modifymember.nolist"], HeadCount.TITLE, HeadCount.VERSION, name))
								end
							end
						else
							-- CURRENT: Player is offline
							moveResult = player:moveToNoList(activityTime, false)
							if moveResult then
								currentRaid:moveToNoGroup(player)
								HeadCount:LogDebug(string.format(L["debug.raid.update.modifymember.offline"], HeadCount.TITLE, HeadCount.VERSION, name))
							end
						end

						player:updateValues(i, className, fileName)	-- update player values if needed
					else
						-- CREATE NEW PLAYER
						-- player is NOT present in the player list (new player)
						local _, race = UnitRace("raid" .. i)
						local sex = UnitSex("raid" .. i)
						local guild = GetGuildInfo("raid" .. i)

						if 0 == level then
							level = UnitLevel("raid" .. i)
						end

						if name then
							-- protect against phantom players
							local player = self:createPlayer(name, race, guild, sex, level, className, fileName, subgroup, isOnline, activityTime)
							currentRaid:addPlayer(player)	-- add the player to the raid player list
						end
					end	-- end raid member is existing present member or new member

					-- track the player
					if name then
						trackedPlayerList[name] = name
					end
				end	-- end loop through all raid members

				-- TODO: Remove/cleanup members that are no longer in the raid.
				currentRaid:processMissingPlayers(numberOfRaidMembers, trackedPlayerList, activityTime)

				-- Update the last activity time for the raid after member processing is fully complete
				currentRaid.lastActivityTime = activityTime
				HeadCount:LogDebug(string.format(L["debug.raid.update.continue"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(activityTime)))
				--HeadCount:LogDebug(string.format(L["debug.raid.update.numberOfRaidMembers"], HeadCount.TITLE, HeadCount.VERSION, currentRaid:getNumberOfPlayers(), HeadCount:getDateTimeAsString(activityTime)))
			else
				-- no raid active, cleanup any existing raids
				if self.isCurrentRaidActive then
					-- current raid is active, end it
					HeadCount:LogDebug(string.format(L["debug.raid.update.activitytime"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(activityTime)))
					self:endCurrentRaid(activityTime)
					HeadCount:LogDebug(string.format(L["debug.raid.update.end"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(activityTime)))
				end
			end	-- end determine number of raid members
		else
			-- cleanup any existing raids
			if self.isCurrentRaidActive then
				-- current raid is active, end it
				HeadCount:LogDebug(string.format(L["debug.raid.update.activitytime"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(activityTime)))
				self:endCurrentRaid(activityTime)
				HeadCount:LogDebug(string.format(L["debug.raid.update.end"], HeadCount.TITLE, HeadCount.VERSION, HeadCount:getDateTimeAsString(activityTime)))
			end
		end -- end determine if mod owner is in a raid or not

		self:SetRaidListWrapper(self.raidListWrapper)	-- save the raid list
	end
end

-- Manage loot updates
-- @param raidId The raid id.
-- @param lootId The loot id.
-- @param playerName The name of the player who looted the item
-- @paam source The loot source
-- @param lootCost The loot cost.
-- @param lootNote The loot note.
function Tracker:lootManagementUpdate(raidId, lootId, playerName, source, lootCost, lootNote)
	assert(raidId, "Unable to process loot management update because the raidId is not found.")
	assert(lootId, "Unable to process loot management update because the lootId is not found.")
	assert(playerName, "Unable to process loot management update because the player name is nil.")
	assert(source, "Unable to process loot management update because the loot source is nil.")
	assert(type(lootCost) == "number", "Unable to process loot management update because the loot cost is not a number.")
	assert(lootNote, "Unable to process loot management update because the loot note is nil.")

	local raid = self.raidListWrapper:getRaidById(raidId)
	if raid then
		local loot = raid:retrieveLoot(lootId)
		if (loot) then
			loot.player = playerName
			loot.source = source
			loot.cost = lootCost
			loot.note = lootNote
			self:SetRaidListWrapper(self.raidListWrapper)	-- save the raid list
		else
			error("Unable to process loot management update because the loot could not be found: " .. lootId)
		end
	else
		error("Unable to process loot management update because the raid could not be found: " .. raidId)
	end
end

-- Creates a player.
-- @param name The name.
-- @param race The race.
-- @param guild The guild name.
-- @param sex The sex.
-- @param level The player level.
-- @param className The localized class name.
-- @param fileName The standard class name.
-- @param subgroup The subgroup.
-- @param isOnline The online status.
-- @param activityTime The activity time.
-- @return object Returns a player or nil if none can be created.
function Tracker:createPlayer(name, race, guild, sex, level, className, fileName, subgroup, isOnline, activityTime)
	local player = nil

	local utcDateTimeInSeconds = HeadCount:getUTCDateTimeInSeconds()
	local activityTime = HeadCount:NewInstance("Time",utcDateTimeInSeconds)

	local timePair = HeadCount:NewInstance("TimePair", { start = activityTime})

	local args = {
		["name"] = name,
		["className"] = className,
		["fileName"] = fileName,
		["race"] = race,
		["guild"] = guild,
		["sex"] = sex,
		["level"] = level,
		["isPresentInRaidList"] = false,
		["isPresentInWaitList"] = false,
		["isOnline"] = true,
		["isWaitlisted"] = false,
		["waitlistActivityTime"] = nil,
		["waitlistNote"] = nil,
		["raidListTime"] = 0,
		["waitListTime"] = 0,
		["offlineTime"] = 0,
		["timeList"] = { },
		["lastActivityTime"] = activityTime,
		["isFinalized"] = false
	}

	if isOnline then
		-- CURRENT: Player is online
		if HeadCount.db.profile.raidListGroups[subgroup] then
			-- CURRENT: Player is in the raid list.
			args["isPresentInRaidList"] = true
			args["isPresentInWaitList"] = false
			args["isOnline"] = true

			timePair:setNote(L["Raid list"])
			table.insert(args["timeList"], timePair)	-- insert the first time pair (includes begin time with no end time)

			player = HeadCount:NewInstance("Player",args)
			HeadCount:LogDebug(string.format(L["debug.raid.update.addmember.raidlist"], HeadCount.TITLE, HeadCount.VERSION, name))
		elseif HeadCount.db.profile.waitListGroups[subgroup] then
			-- CURRENT: Player is in the wait list.
			args["isPresentInRaidList"] = false
			args["isPresentInWaitList"] = true
			args["isWaitlisted"] = true
			args["waitlistActivityTime"] = activityTime
			args["isOnline"] = true

			timePair:setNote(L["Wait list"])
			table.insert(args["timeList"], timePair)	-- insert the first time pair (includes begin time with no end time)

			player = HeadCount:NewInstance("Player",args)
			HeadCount:LogDebug(string.format(L["debug.raid.update.addmember.waitlist"], HeadCount.TITLE, HeadCount.VERSION, name))
		else
			-- CURRENT: Player is in no list.
			args["isPresentInRaidList"] = false
			args["isPresentInWaitList"] = false
			args["isOnline"] = true

			timePair:setNote(L["No list"])
			table.insert(args["timeList"], timePair)	-- insert the first time pair (includes begin time with no end time)

			player = HeadCount:NewInstance("Player",args)
			HeadCount:LogDebug(string.format(L["debug.raid.update.addmember.nolist"], HeadCount.TITLE, HeadCount.VERSION, name))
		end
	else
		-- CURRENT: Player is offline
		args["isPresentInRaidList"] = false
		args["isPresentInWaitList"] = false
		args["isOnline"] = false

		timePair:setNote(L["Offline"])
		table.insert(args["timeList"], timePair)	-- insert the first time pair (includes begin time with no end time)

		player = HeadCount:NewInstance("Player",args)
		HeadCount:LogDebug(string.format(L["debug.raid.update.addmember.offline"], HeadCount.TITLE, HeadCount.VERSION, name))
	end

	return player
end

----------------------------
-- BOSS KILL TRACKING
----------------------------

-- Determines if a boss is engaged.
-- @return boolean Returns true if boss is engaged and false otherwise
-- @return boolean Returns true if the boss engage check needs to be checked again and false otherwise.
function Tracker.fixme:processBossEngagement()
	local isBossEngaged = false
	local isRecheckNeeded = false

	if self.isCurrentRaidActive then
		-- only check if boss is engaged if in an active raid
		local bossTable = self:processBossTargets()

		if bossTable then
			for k,v in pairs(bossTable) do
				local guid = v["guid"]
				local isDead = v["isDead"]
				if not isDead and not self.currentBossEvent:isPresent(guid) then
					-- boss is engaged, alive and boss is not present in the event tracker, add it!
					HeadCount:LogDebug(string.format(L["debug.boss.engage"], HeadCount.TITLE, HeadCount.VERSION, v["bossName"], guid))

					self.currentBossEvent:addBoss(v)	-- add the boss to the boss event

					isBossEngaged = true
				end
			end

			if isBossEngaged then
				--HeadCount:removeEvent("PLAYER_REGEN_DISABLED")		-- disable player enter combat event
				--HeadCount:addEvent("PLAYER_REGEN_ENABLED")			-- enable player leave combat event
				--HeadCount:addEvent("COMBAT_LOG_EVENT_UNFILTERED")	-- add combat log event
				--HeadCount:addEvent("CHAT_MSG_MONSTER_YELL")			-- add monster yell event
			else
				-- boss is not engaged yet
				if UnitAffectingCombat("player") then
					-- recheck for boss engagement
					isRecheckNeeded = true
				end
			end
		end
	else
		-- no raid is active, disable all boss kill tracking events
		self:endRaid()
	end

	return isBossEngaged, isRecheckNeeded
end

-- Processes boss targets
-- @return table Returns the boss table.
-- @return number Returns the number of entries in the boss table.
function Tracker:processBossTargets()
	local bossTable = { }
	local numberOfBosses = 0

	-- the current zone
	local zone = GetRealZoneText()
	local difficulty = HeadCount:determineDifficulty()
	local isBoss = false
	local guid = nil
	local bossName = nil
	local isDead = nil

	isBoss, guid, bossName, isDead = self:isBossUnit("target")
	if isBoss then
		-- target exists, target is in combat, target is a boss mob for this zone
		bossTable[guid] = { ["guid"] = guid, ["bossName"] = bossName, ["zone"] = zone, ["isDead"] = isDead, }
		numberOfBosses = numberOfBosses + 1
	end

	isBoss, guid, bossName, isDead = self:isBossUnit("focus")
	if isBoss then
		-- focus exists, focus is in combat, focus is a boss mob for this zone
		if not bossTable[guid] then
			bossTable[guid] = { ["guid"] = guid, ["bossName"] = bossName, ["zone"] = zone, ["isDead"] = isDead, }
			numberOfBosses = numberOfBosses + 1
		end
	end

	local numberOfRaidMembers = GetNumRaidMembers()
	if numberOfRaidMembers > 0 then
		for i = 1, numberOfRaidMembers do
			local raidUnit = string.format("%s%d%s", "raid", i, "target")
			isBoss, guid, bossName, isDead = self:isBossUnit(raidUnit)
			if isBoss then
				-- raid unit target exists, raid unit target is in combat, raid unit target is a boss mob for this zone
				if not bossTable[guid] then
					bossTable[guid] = { ["guid"] = guid, ["bossName"] = bossName, ["zone"] = zone, ["difficulty"] = difficulty, ["status"] = isDead, }
					numberOfBosses = numberOfBosses + 1
				end
			end
		end
	end

	return bossTable, numberOfBosses
end

-- Determines if the current unit is a valid boss in combat
-- @param unit The target unit
-- @return boolean Returns true if the unit is a valid boss and false otherwise.
-- @return string Returns the guid.
-- @return string Returns the boss name or nil if the unit is not a valid boss.
-- @return boolean Returns if the unit is dead or alive
function Tracker.fixme:isBossUnit(unit)
	local isValid = false
	local guid = nil
	local name = nil
	local isDead = nil
	if unit then
		if UnitExists(unit) then
			-- unit exists
			local unitName = UnitName(unit)
			if UnitAffectingCombat(unit) then
				if UnitClassification(unit) == "worldboss" then
					local alias = HeadCount.BOSS_ALIASES[unit]
					if InstanceBossList[unitName] or InstanceBossList[alias] then
						isValid = true
						name = alias or unitName
						isDead = GetInstanceLockTimeRemainingEncounter(InstanceBossList[name])
					end
				end
			end

			--if UnitAffectingCombat(unit) and UnitClassification(unit) == "worldboss" and not HeadCount.BOSS_BLACKLIST[unitName] then
			--	-- boss is in combat, boss is a worldboss, boss is not on the ignore list, boss is not tracked yet
			--	isValid = true
			--	guid = self:retrieveGUID(unit)
			---	name = unitName
			--	isDead = UnitIsDead(unit)
			--elseif (UnitAffectingCombat(unit) and (HeadCount.BOSS_WHITELIST[unitName])) then
			--	-- boss is in combat, boss is on the whitelist, boss is not tracked yet
			--	isValid = true
			--	guid = self:retrieveGUID(unit)
			--	name = unitName
			--	isDead = UnitIsDead(unit)
			--end
		end
	end
	return isValid, guid, name, isDead
end

-- Gets the unit GUID.
-- @param unit The unit.
-- @return string Returns the unit GUID or nil if none exists
function Tracker:retrieveGUID(unit)
	local guid = nil

	if unit then
		guid = UnitGUID(unit)
		if 0 == tonumber(guid, 16) then
			return
		end
	end
	return guid
end

-- Determines if the raid has wiped on a boss
-- @return boolean Returns true if there is a boss wipe and false otherwise
-- @return boolean Returns true if the boss wipe check needs to be checked again and false otherwise.
function Tracker.fixme:processBossWipe()
	local isBossWipe = false
	local isRecheckNeeded = false

	if self.isCurrentRaidActive then
		if self.currentBossEvent.started then
			-- only check for boss wipe if in an active raid an a boss is engaged
			HeadCount:LogDebug(string.format(L["debug.boss.wipe.check"], HeadCount.TITLE, HeadCount.VERSION))

			if not UnitIsFeignDeath("player") then
				-- get the current target list
				local _, numberOfBosses = self:processBossTargets()
				if (numberOfBosses > 0) then
					-- alive bosses are engaged by the raid, recheck for wipe!
					isRecheckNeeded = true
				else
					-- no alive bosses are engaged by the raid, it's a wipe!
					-- no boss is engaged
					--HeadCount:removeEvent("CHAT_MSG_MONSTER_YELL")			-- disable monster yell events
					--HeadCount:removeEvent("COMBAT_LOG_EVENT_UNFILTERED")	-- disable combat log event
					--HeadCount:removeEvent("PLAYER_REGEN_ENABLED")			-- disable player leave combat event
					--HeadCount:addEvent("PLAYER_REGEN_DISABLED")				-- enable player enter combat event

					-- no boss is engaged, raid has wiped
					HeadCount:LogDebug(string.format(L["debug.boss.wipe"], HeadCount.TITLE, HeadCount.VERSION))
					-- destroy current boss event
					self.currentBossEvent:Stop()
					isBossWipe = true
				end
			end
		else
			-- no boss event is currently active, reset to standard event handling
			--HeadCount:removeEvent("CHAT_MSG_MONSTER_YELL")			-- disable monster yell events
			--HeadCount:removeEvent("COMBAT_LOG_EVENT_UNFILTERED")	-- disable combat log event
			--HeadCount:removeEvent("PLAYER_REGEN_ENABLED")			-- disable player leave combat event
			--HeadCount:addEvent("PLAYER_REGEN_DISABLED")				-- enable player enter combat event
			self.currentBossEvent:Stop()
		end
	else
		-- no raid is active, disable all boss kill tracking events
		self:endRaid()
	end

	return isBossWipe, isRecheckNeeded
end

-- Determines if a boss has died
-- @param guid The mob unit guid.
-- @param mob The mob that has just died.
-- @return boolean Returns true if a boss kill event is complete and false otherwise.
function Tracker.fxime:processBossDeath(guid, mob)
	local isBossKillEventProcessed = false

	if self.isCurrentRaidActive then
		if self.currentBossEvent.started then
			if not guid and mob then
				-- try and look up the guid from the mob
				guid = self.currentBossEvent:getGUID(mob)
			end
			if not guid then return isBossKillEventProcessed end

			local isPresent = self.currentBossEvent:isPresent(guid)
			local isAlive = self.currentBossEvent:isAlive(guid)
			if isPresent and isAlive then
				-- boss kill recorded!
				local encounterName = self.currentBossEvent.encounter
				HeadCount:LogDebug(string.format(L["debug.boss.kill"], HeadCount.TITLE, HeadCount.VERSION, self.currentBossEvent:getName(guid), guid))
				self.currentBossEvent:AddKill(guid)

				-- check for any new targets
				self:processBossEngagement()
				if (self.currentBossEvent:Complete()) then
					-- boss event is complete!
					HeadCount:removeEvent("CHAT_MSG_MONSTER_YELL")			-- disable monster yell events
					HeadCount:removeEvent("COMBAT_LOG_EVENT_UNFILTERED")	-- disable combat log event
					HeadCount:removeEvent("PLAYER_REGEN_ENABLED")			-- disable player leave combat event
					HeadCount:addEvent("PLAYER_REGEN_DISABLED")				-- enable player enter combat event

					local utcDateTimeInSeconds = HeadCount:getUTCDateTimeInSeconds()
					local activityTime = HeadCount:NewInstance("Time",utcDateTimeInSeconds)

					local zone = GetRealZoneText()
					local difficulty = HeadCount:determineDifficulty()

					local isBossKillProcessed = self:addBossKill(encounterName, zone, difficulty, activityTime)	-- add the boss kill
					if (isBossKillProcessed) then
						isBossKillEventProcessed = true
					end

					self.currentBossEvent:Stop()
				end
			end
		else
			-- no boss event is currently active, reset to standard event handling
			--HeadCount:removeEvent("CHAT_MSG_MONSTER_YELL")			-- disable monster yell events
			--HeadCount:removeEvent("COMBAT_LOG_EVENT_UNFILTERED")	-- disable combat log event
			--HeadCount:removeEvent("PLAYER_REGEN_ENABLED")			-- disable player leave combat event
			--HeadCount:addEvent("PLAYER_REGEN_DISABLED")				-- enable player enter combat event

			self.currentBossEvent:Stop()
		end
	else
		-- no raid is active, disable all boss kill tracking events
		self:endRaid()
	end

	return isBossKillEventProcessed
end

-- Prunes old raids.
function Tracker:pruneRaids()
	local isPruningEnabled = HeadCount.db.profile.raid["pruning"]["isPruningEnabled"]
	local pruningTimeInWeeks = HeadCount.db.profile.raid["pruning"]["pruningTime"]

	if (isPruningEnabled) then
		HeadCount:LogDebug(string.format(L["debug.raid.prune.enabled"], HeadCount.TITLE, HeadCount.VERSION, pruningTimeInWeeks))

		self.raidListWrapper:pruneRaids(pruningTimeInWeeks)	-- prune old raids
	else
		HeadCount:LogDebug(string.format(L["debug.raid.prune.disabled"], HeadCount.TITLE, HeadCount.VERSION))
	end
end

-- Processes an incoming message.
-- @param message The message.
-- @param author The author.
-- @return boolean Returns true if an update is required as a result of the message.
-- @return boolean Returns true if the message is a supported message.
-- @return table Returns the response message list.
function Tracker:processIncomingMessage(message, author)
	local isUpdateRequired = false
	local isSupportedMessage = false
	local messageList = nil

	if self.isCurrentRaidActive then
		-- Only process incoming messages during an active raid
		local raid = self:retrieveCurrentRaid()
		isUpdateRequired, isSupportedMessage, messageList = self.messageHandler:processIncomingMessage(raid, self.isWaitlistAcceptanceEnabled, message, author)
	end

	return isUpdateRequired, isSupportedMessage, messageList
end

-- Determines if an incoming message is supported.
-- @param message The message.
-- @return boolean Returns true if the message is a supported message.
function Tracker:IsSupportedIncomingMessage(message)
	assert(message, "Unable to determine if incoming message is supported because the message is nil.")
	local isSupportedMessage = false
	if self.isCurrentRaidActive then
		isSupportedMessage = self.messageHandler:IsSupportedIncomingMessage(message)
	end
	return isSupportedMessage
end

-- Determines if an outgoing message is supported.
-- @param message The message.
-- @return boolean Returns true if the message is a supported message.
function Tracker:IsSupportedOutgoingMessage(message)
	assert(message, "Unable to determine if the outgoing message is supported because the message is nil.")
	local isSupportedMessage = false
	if self.isCurrentRaidActive then
		isSupportedMessage = self.messageHandler:IsSupportedOutgoingMessage(message)
	end
	return isSupportedMessage
end

function Tracker:processListWaitlist(channel)
	if self.isCurrentRaidActive then
		local raid = self:retrieveCurrentRaid()
		self.messageHandler:processListWaitlistMessage(raid, channel)
	else
		HeadCount:LogWarning(string.format(L["warning.waitlist.list.currentraid"], HeadCount.TITLE, HeadCount.VERSION))
	end
end

--[[
	Event Handlers
--]]
function Tracker:CHAT_MSG_WHISPER(...)
end
function Tracker:PLAYER_REGEN_DISABLED()
end
function Tracker:CHAT_MSG_MONSTER_YELL(...)
end
function Tracker:COMBAT_LOG_EVENT_UNFILTERED(...)
end
--[[
local function BuildSavedInfoList()
	for k,v in pairs(GetFullRaidList()) do
	local 	name, id, reset, difficulty, locked, extended, instanceIDMostSig, isRaid, maxPlayers, difficultyName, numEncounters, encounterProgress = GetSavedInstanceInfo(k)
		InstanceList[k] = name
		for m=1,numEncounters do
			local  bossName,text,isKilled = GetSavedInstanceEncounterInfo(k,m)
			BossList[bossName] = k
		end
	end
end

local function DetermineInstance()
	local lockTimeleft, isPreviousInstance, encountersTotal, encountersComplete = GetInstanceLockTimeRemaining();
	local bossName, texture, isKilled = GetInstanceLockTimeRemainingEncounter(i);
	CurrInstance = BossList[bossName]
end

local function ScanInstanceList()
	if CurrInstance and CurrInstance > 0 then
		local lockTimeleft, isPreviousInstance, encountersTotal, encountersComplete = GetInstanceLockTimeRemaining();
		for i=1,encountersTotal do
			local  bossName,text,isKilled = GetSavedInstanceEncounterInfo(CurrInstance,i)
			KilledBosses[bossName] = isKilled
		end
	end
end

BuildSavedInfoList()

local eventFrame = CreateFrame("Frame","KgUIEventFrame")
eventFrame:RegisterEvent("UPDATE_INSTANCE_INFO")
eventFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
eventFrame:SetScript("OnEvent", function(frame,event,...)

	if event == "UPDATE_INSTANCE_INFO" then
		if CurrInstance == 0 or CurrInstance == nil then
			BuildSavedInfoList()
			DetermineInstance()
			ScanInstanceList()
		end
		print("Instance Data updated, checking for Boss credit.")
		local tmp = {}
		for k,v in pairs(KilledBosses) do
			tmp[k] = v
		end
		ScanInstanceList()
		for k,v in pairs(KilledBosses) do
			if v ~= tmp[k] then
				print("You killed ",k)
			end
		end
	end
	if event == "PLAYER_REGEN_ENABLED" then
		RequestRaidInfo()
	end
end)

--]]