local HeadCount = LibStub("AceAddon-3.0"):GetAddon("HeadCount2")
if not HeadCount then return end
local L = LibStub("AceLocale-3.0"):GetLocale("HeadCount2")

-- Converts a string to a proper name (first letter capitalized, remainder of string in lower case).
-- @param originalString The original string to convert.
function HeadCount:convertStringToProperName(originalString)
	return string.gsub(string.lower(originalString), "%a", string.upper, 1)	-- string in all lowercase except for first letter in uppercase
end

-- Splits a string by a given pattern.
-- @param originalString The original string.
-- @param pattern The delimiter pattern.
-- @return table Returns a table of the split string.
function HeadCount:split(originalString, pattern)
	local temp = {}
	local fPattern = "(.-)" .. pattern
	local lastEnd = 1
	local s, e, cap = originalString:find(fPattern, 1)
	while s do
		if ((s ~= 1) or (cap ~= "")) then
			table.insert(temp, cap)
		end
		lastEnd = e + 1
		s, e, cap = originalString:find(fPattern, lastEnd)
	end
	if (lastEnd <= # originalString) then
		cap = originalString:sub(lastEnd)
		table.insert(temp, cap)
	end
	return temp
end

function HeadCount:getUTCDateTimeInSeconds()
	local utcDateTime = date("!*t")
	return time(utcDateTime)
end

function HeadCount:getDateTimeAsString(activityTime)
	local currentSeconds = activityTime.epoc	-- UTC date and time in seconds
	local displayTimezoneDifferenceInSeconds = self.db.profile.timezone * 60 * 60 -- seconds
	currentSeconds = currentSeconds + displayTimezoneDifferenceInSeconds			 -- display time in seconds
	local displayDateTime = date("*t", currentSeconds)
	local formattedDate = self:formatDate(displayDateTime)	-- retrieve the formatted date
	local formattedTime = self:formatTime(displayDateTime) -- retrieve the formatted time
	local datetimeString = nil
	if ((formattedDate) and (formattedTime)) then
		datetimeString = formattedDate .. " " .. formattedTime
	else
		error("Unable to process date time string because the formatted date and/or time are not initialized.")
	end
	return datetimeString
end

function HeadCount:formatTime(dateTime)
	local formattedTime = nil
	if (dateTime) then
		local isTimeFormatEnabled = HeadCount.db.profile.isTimeFormatEnabled
		if (isTimeFormatEnabled) then
			-- 24 hour clock
			formattedTime = string.format("%02d", dateTime.hour) .. ":" .. string.format("%02d", dateTime.min) .. ":" .. string.format("%02d", dateTime.sec)
		else
			-- 12 hour clock
			local hour = dateTime.hour
			if (dateTime.hour <= 11) then
				-- 00:00 - 11:59  (am)
				if (0 == dateTime.hour) then
					hour = dateTime.hour + 12
				end

				formattedTime = string.format("%d", hour) .. ":" .. string.format("%02d", dateTime.min) .. ":" .. string.format("%02d", dateTime.sec) .. L["am"]
			else
				-- 12:00 - 23:59 (pm)
				if (dateTime.hour >= 13) then
					hour = dateTime.hour - 12
				end
				formattedTime = string.format("%d", hour) .. ":" .. string.format("%02d", dateTime.min) .. ":" .. string.format("%02d", dateTime.sec) .. L["pm"]
			end
		end
	end
	return formattedTime
end

function HeadCount:formatDate(dateTime)
	local formattedDate = nil
	if (dateTime) then
		local dateFormat = HeadCount:GetDateFormat()
		if (1 == dateFormat) then
			formattedDate = dateTime.day .. "/" .. dateTime.month .. "/" .. dateTime.year
		elseif (2 == dateFormat) then
			formattedDate = dateTime.month .. "/" .. dateTime.day .. "/" .. dateTime.year
		else
			formattedDate = dateTime.year .. "/" .. dateTime.month .. "/" .. dateTime.day
		end
	end
	return formattedDate
end

function HeadCount:getDateAsString(currentTime)
	-- TODO: Add locale to this method to display time in various localized formats
	local currentSeconds = currentTime.epoc
	local displayTimezoneDifferenceInSeconds = self.db.profile.timezone * 60 * 60 -- seconds
	currentSeconds = currentSeconds + displayTimezoneDifferenceInSeconds			 -- display time in seconds
	local displayDateTime = date("*t", currentSeconds)
	local formattedDate = self:formatDate(displayDateTime)	-- get the formatted date
	return formattedDate
end

function HeadCount:getTimeAsString(activityTime)
	-- TODO: Add locale to this method to display time in various localized formats
	local timeString = "00:00:00"
	if (activityTime) then
		local currentSeconds = activityTime.epoc		-- UTC date and time in seconds
		local displayTimezoneDifferenceInSeconds = self.db.profile.timezone * 60 * 60 -- seconds
		currentSeconds = currentSeconds + displayTimezoneDifferenceInSeconds			 -- display time in seconds
		local displayDateTime = date("*t", currentSeconds)
		local formattedTime = self:formatTime(displayDateTime)
		if (formattedTime) then
			timeString = formattedTime
		else
			error("Unable to process date time string because the formatted time is not initialized.")
		end
	end
	return timeString
end

function HeadCount:getSecondsAsString(totalSeconds)
	local secondsString = nil
	local remainingSeconds
	if (totalSeconds > 0) then
		local numberOfHours = math.floor(totalSeconds / 3600)
		remainingSeconds = totalSeconds - (numberOfHours * 3600)
		local numberOfMinutes = math.floor(remainingSeconds / 60)
		remainingSeconds = remainingSeconds - (numberOfMinutes * 60)
		secondsString = string.format("%02d:%02d:%02d", numberOfHours, numberOfMinutes, remainingSeconds)
	else
		secondsString = "00:00:00"
	end
	return secondsString
end

-- Computes the time difference between a current date and an original date.
-- @param currentTime The current time.
-- @param originalTime The original time.
-- @return number Returns the time difference in seconds between the current date and an original date.
function HeadCount:computeTimeDifference(currentTime, originalTime)
	local timeDifference = 0
	if ((currentTime) and (originalTime)) then
		local currentSeconds = currentTime.epoc
		local originalSeconds = originalTime.epoc
		timeDifference = difftime(currentSeconds, originalSeconds)
	end
	return timeDifference
end

-- Converts a boolean value to its localized string equivalent.
-- @param booleanValue The boolean value.
-- @return string Returns "true" if the boolean value is true, "false" otherwise
function HeadCount:convertBooleanToString(booleanValue)
	if (booleanValue) then
		-- boolean value is not nil -> true
		return L["True"]
	else
		-- boolean value is nil -> false
		return L["False"]
	end
end

-- Converts a boolean value to its localized string equivalent.
-- @param booleanValue The boolean value.
-- @return string Returns "true" if the boolean value is true, "false" otherwise
function HeadCount:convertBooleanToYesNoString(booleanValue)
	if (booleanValue) then
		-- boolean value is not nil -> true
		return L["Yes"]
	else
		-- boolean value is nil -> false
		return L["No"]
	end
end

function HeadCount:isRaidInstance()
	local isPresent = false
	local isPresentInInstance, instanceType = IsInInstance()
	if ((isPresentInInstance) and (instanceType == "raid")) then
		-- player is in a raid instance
		isPresent = true
	end
	return isPresent
end

-- Determines the boss encounter name
-- @param name The boss name
-- @return encounterName Returns the boss encounter name
function HeadCount:retrieveBossEncounterName(name)
	assert(name, "Unable to retrieve boss encounter name because the name is nil.")
	local encounterName = nil
	-- BABBLE BOSS STUFF HERE
	if (HeadCount.BOSS_ALIASES[name]) then
		-- special boss encounter name
		encounterName = HeadCount.BOSS_ALIASES[name]
	else
		-- encounter name is the boss
		encounterName = name
	end
	return encounterName
end

-- Determines if the zone is a battleground/arena zone
function HeadCount:isBattlegroundZone(zone)
	assert(zone, "Unable to determine if zone is a battleground because the zone is nil.")
	local isBattleground = false
	local inInstance, instanceType = IsInInstance()
	if ((instanceType == "pvp") or (instanceType == "arena")) then
		isBattleground = true
	end
	return isBattleground
end

-- Determines the difficulty level
-- 1 - 10 Player
-- 2 - 25 Player
-- 3 - 10 Player (Heroic)
-- 4 - 25 Player (Heroic)
-- @return number Returns the raid difficulty level or nil if not in a raid instance
function HeadCount:determineDifficulty()
	local difficulty = nil
	local inInstance, instanceType = IsInInstance()
	if (inInstance and instanceType == "raid") then
		difficulty = GetRaidDifficulty()
	end
	return difficulty
end

-- Escape characters into XML entities
-- @param value The string to convert
-- @return string Returns the converted entity string
function HeadCount:convertToXMLEntities(value)
	local xmlString = nil
	if (value) then
		xmlString = value
        xmlString = string.replace(xmlString, "&", "&amp;" )
        xmlString = string.replace(xmlString,  "<", "&lt;" )
        xmlString = string.replace(xmlString,  ">", "&gt;" )
        xmlString = string.replace(xmlString,  "\"", "&quot;" )
        xmlString = string.replace(xmlString,  "'", "&apos;" )
	end
	return xmlString
end

-- Converts a time to a W3C international standard date and time string  (ISO 8601)
-- @param activityTime The activity time.
-- @return string Returns the display date and time as a ISO 8601 string
function HeadCount:getDateTimeAsXMLString(activityTime)
   local dateTimeString = nil
	if (activityTime) then
		local currentSeconds = activityTime:getUTCDateTimeInSeconds()      -- UTC date and time in seconds
		local displayDateTime = date("*t", currentSeconds)
		-- Example: July 7, 2008 3:46:42 UTC is equivalent to 2008-07-07T03:46:42Z
		dateTimeString =  string.format("%04d", displayDateTime.year) .. "-" .. string.format("%02d", displayDateTime.month) .. "-" .. string.format("%02d", displayDateTime.day) .. "T" .. string.format("%02d", displayDateTime.hour) .. ":" .. string.format("%02d", displayDateTime.min) .. ":" .. string.format("%02d", displayDateTime.sec) .. "Z"
	end
	return dateTimeString
end

local ITEM_COLORS_NO_ALPHA = { "9D9D9D", "FFFFFF", "1EFF00", "0070DD", "A335EE", "FF8000", "E6CC80", }
local DEFAULT_COLOR_NO_ALPHA = "A335EE"

function HeadCount:retrieveItemColor(rarity)
	local color = DEFAULT_COLOR_NO_ALPHA

	if rarity then
		if rarity >= 0 and rarity <= 6 then
			local lookupRarity = rarity + 1
			color = ITEM_COLORS_NO_ALPHA[lookupRarity]
		end
	end

	return color
end


function HeadCount:RunExporterTest()
	print("Starting Exporter Test")
	local activityTime = self:NewInstance("Time",HeadCount:getUTCDateTimeInSeconds())
	local args = {
		["playerList"] = {
			["raidgroup"] = { },
			["nogroup"] = { },
		},
		["bossList"] = { },
		["lootList"] = { },
		["timeList"] = { },
		["difficulty"] = nil,
		["raidTime"] = 0,
		["lastActivityTime"] = activityTime,
		["zone"] = "Unknown",
		["numberOfPlayers"] = 0,
		["numberOfBosses"] = 0,
		["isFinalized"] = false
	}
	print("Created new raid")
	local raid = self:NewInstance("Raid",args)
	-- Create a Fake Player
	local utcDateTimeInSeconds = HeadCount:getUTCDateTimeInSeconds()


	local timePair = self:NewInstance("TimePair")
	timePair.start = activityTime
	tinsert(raid.timeList,timePair)
	args = {
		["name"] = "Bogus",
		["className"] = "DeathKnight",
		["fileName"] = "DEATHKNIGHT",
		["race"] = "Night Elf",
		["guild"] = "NoGuild",
		["sex"] = 2,
		["level"] = 80,
		["isPresentInRaidList"] = true,
		["isPresentInWaitList"] = false,
		["isOnline"] = true,
		["isWaitlisted"] = false,
		["waitlistActivityTime"] = nil,
		["waitlistNote"] = nil,
		["raidListTime"] = 0,
		["waitListTime"] = 0,
		["offlineTime"] = 0,
		["timeList"] = { },
		["lastActivityTime"] = activityTime,
		["isFinalized"] = false
	}
	print("Creted player")
	local player = self:NewInstance("Player",args)
	print("Creating player complete")
	tinsert(player.timeList,timePair)
	print("Added raid player")
	raid:addPlayer(player)
	print("Calling Exporter")
	raid:finalize(activityTime)
	local result = self:exportRaid(raid)
	print(result)
	print("EXPORT COMPLETE")
end